@@ -555,7 +555,13 @@ ascii_art = {
 ###################################################################################
 
 
-def print_art(*art_names, all=False, suffix=None, output_file=None, output_path=None):
+def print_art(
+    *art_names,
+    all=False,
+    suffix=None,
+    output_file=None,
+    output_path=None,
+):
     """
     Print ASCII art based on user input and optionally save output to a
     specified path.
